/*  This file contains or coordinates all of the code necessary to compile
    on a UNIX system that can use svgalib, such as FreeBSD and Linux.  
    This code is not guaranteed to compile on FreeBSD, though.
*/


#include <unistd.h>
#include <sys/types.h>
#include <signal.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <strings.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <assert.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <sched.h>
#include <vga.h>
#include <vgakeyboard.h>
#include <vgamouse.h>
#ifdef LINUX
#include <linux/joystick.h>
#endif

#include "../../driver.h"

#include "video.h"

#include "../common/config.h"
#include "../common/args.h"
#include "../common/unixdsp.h"
#include "../common/cheat.h"

static int ntsccol=0,ntschue=-1,ntsctint=-1;

static char netplayhost[256];
static int netskip;
static int netplay=0;
static int f8bit=0;
static int soundvol=100;
static char BaseDirectory[2048];
#define EMO_NO8LIM      1
#define EMO_SUBASE      2

int FCEUDemuloptions=0;

static int winit=0;
static char TempArray[2048];

static void DriverKill(void);
static int DriverInitialize(void);

static int soundo=48000;
static int sfragsize=7,snfrags=8;

static int gametype;
void FCEUD_SendMessage(int msg)
{
 switch(msg)
 {
  case DSM_RESETSPECIAL: gametype=0;break;
  case DSM_SPECIALSYSTEM_VSUNISYSTEM:
  case DSM_SPECIALSYSTEM_NSF:
  case DSM_SPECIALSYSTEM_FDS: gametype=msg;break;
 }
}

void FCEUD_PrintError(char *s)
{
 puts(s);
}

#include "input.c"
#include "netplay.c"

static char *cpalette=0;
static void LoadCPalette(void)
{
 char tmpp[192];
 FILE *fp;

 if(!(fp=fopen(cpalette,"rb")))
 {
  printf(" Error loading custom palette from file: %s\n",cpalette);
  return;
 }
 fread(tmpp,1,192,fp);
 FCEUI_SetPaletteArray(tmpp);
 fclose(fp);
}

#define CONFIGITEMS 19
static CFGSTRUCT fceuconfig[CONFIGITEMS]={
	{"soundvol",&soundvol,sizeof(soundvol)},
	{"cpalette",&cpalette,0},
	{"snfrags",&snfrags,sizeof(snfrags)},
	{"sfragsize",&sfragsize,sizeof(sfragsize)},
        {"sound",&soundo,sizeof(soundo)},
        {"vmode",&FCEUDvmode,sizeof(FCEUDvmode)},
        {"joybmap",joyBMap,sizeof(joyBMap)},
        {"joy",joy,sizeof(joy)},
        {"ntsctint",&ntsctint,sizeof(ntsctint)},
        {"ntschue",&ntschue,sizeof(ntschue)},
        {"ntsccol",&ntsccol,sizeof(ntsccol)},
        {"f8bit",&f8bit,sizeof(f8bit)},
	{"joymulti",&joymulti,sizeof(joymulti)},
        {"InputType",InputType,sizeof(InputType)},
        {"powerpadside",&powerpadside,sizeof(powerpadside)},
        {"powerpadsc",powerpadsc,sizeof(powerpadsc)},
        {"FCEUDemuloptions",&FCEUDemuloptions,sizeof(FCEUDemuloptions)},
        {"srendline",&FCEUDsrendline,sizeof(FCEUDsrendline)},
        {"erendline",&FCEUDerendline,sizeof(FCEUDerendline)}
};

static void SaveConfig(void)
{	
	sprintf(TempArray,"%s/fceu.cfg",BaseDirectory);
        DriverInterface(DES_GETNTSCTINT,&ntsctint);
        DriverInterface(DES_GETNTSCHUE,&ntschue);
        SaveFCEUConfig(TempArray,fceuconfig,CONFIGITEMS);
}

static void LoadConfig(void)
{
        sprintf(TempArray,"%s/fceu.cfg",BaseDirectory);
        LoadFCEUConfig(TempArray,fceuconfig,CONFIGITEMS);
        if(ntsctint>=0) DriverInterface(DES_SETNTSCTINT,&ntsctint);
        if(ntschue>=0) DriverInterface(DES_SETNTSCHUE,&ntschue);
}

static void CreateDirs(void)
{
 mkdir(BaseDirectory,S_IRWXU);
 sprintf(TempArray,"%s/fcs",BaseDirectory);
 mkdir(TempArray,S_IRWXU);
 sprintf(TempArray,"%s/snaps",BaseDirectory);
 mkdir(TempArray,S_IRWXU);
 sprintf(TempArray,"%s/gameinfo",BaseDirectory);
 mkdir(TempArray,S_IRWXU);
 sprintf(TempArray,"%s/sav",BaseDirectory);
 mkdir(TempArray,S_IRWXU);
 sprintf(TempArray,"%s/cheats",BaseDirectory);
 mkdir(TempArray,S_IRWXU);
}

static void GetBaseDirectory(void)
{
 char *ol;
 ol=getenv("HOME");
 BaseDirectory[0]=0;
 if(ol) 
  {
   strncpy(BaseDirectory,ol,2047);
   BaseDirectory[2047]=0;
   strcat(BaseDirectory,"/.fceultra");
   FCEUI_SetBaseDirectory(BaseDirectory);
  }
}

static void SetSignals(void (*t)(int))
{
  int sigs[11]={SIGINT,SIGTERM,SIGHUP,SIGPIPE,SIGSEGV,SIGFPE,SIGKILL,SIGALRM,SIGABRT,SIGUSR1,SIGUSR2};
  int x;
  for(x=0;x<11;x++)
   signal(sigs[x],t);
}

static void CloseStuff(int signum)
{
	DriverKill();
        printf("\nSignal %d has been caught and dealt with...\n",signum);
        switch(signum)
        {
         case SIGINT:printf("How DARE you interrupt me!\n");break;
         case SIGTERM:printf("MUST TERMINATE ALL HUMANS\n");break;
         case SIGHUP:printf("Reach out and hang-up on someone.\n");break;
         case SIGPIPE:printf("The pipe has broken!  Better watch out for floods...\n");break;
         case SIGSEGV:printf("Iyeeeeeeeee!!!  A segmentation fault has occurred.  Have a fluffy day.\n");break;
         case SIGBUS:printf("I told you to be nice to the driver.\n");break;
         case SIGFPE:printf("Those darn floating points.  Ne'er know when they'll bite!\n");break;
         case SIGALRM:printf("Don't throw your clock at the meowing cats!\n");break;
         case SIGABRT:printf("Abort, Retry, Ignore, Fail?\n");break;
         case SIGUSR1:
         case SIGUSR2:printf("Killing your processes is not nice.\n");break;
        }
        exit(1);
}

static void DoArgs(int argc, char *argv[])
{
        static char *cortab[4]={"none","gamepad","zapper","powerpad"};
        static int cortabi[4]={DII_SETCONTROL_NONE,DII_SETCONTROL_GAMEPAD,
                               DII_SETCONTROL_ZAPPER,DII_SETCONTROL_POWERPAD};
	int x;
	static char inputa[2][32];
	static int docheckie[6];

	static ARGPSTRUCT FCEUArgs[24]={
	 {"-soundvol",0,&soundvol,0},
	 {"-cpalette",0,&cpalette,0x4001},

         {"-vmode",0,&FCEUDvmode,0},{"-ntsccol",0,&ntsccol,0},
         {"-pal",&docheckie[0],0,0},{"-sound",0,&soundo,0},
	 {"-snfrags",0,&snfrags,0},{"-sfragsize",0,&sfragsize,0},
         {"-f8bit",0,&f8bit,0},
	 {"-joymulti",0,&joymulti,0},
         {"-input1",0,inputa[0],1},{"-input2",0,inputa[1],1},

	 {"-joy1",0,&joy[0],0},{"-joy2",0,&joy[1],0},
         {"-joy3",0,&joy[2],0},{"-joy4",0,&joy[3],0},

	 {"-gg",&docheckie[1],0,0},
	 {"-no8lim",0,&FCEUDemuloptions,0x8001},{"-subase",&docheckie[3],&FCEUDemuloptions,0x8002},
	 {"-vsync",0,&FCEUDemuloptions,0x8004},
	 {"-slstart",0,&FCEUDsrendline,0},{"-slend",0,&FCEUDerendline,0},
 	 {"-connect",&docheckie[4],netplayhost,1},{"-server",&docheckie[5],&netskip,0},
	};

        inputa[0][0]=inputa[1][0]=0;
        memset(docheckie,0,sizeof(docheckie));

	ParseArguments(argc, argv, FCEUArgs, 24);
	if(cpalette)
	{
  	 if(cpalette[0]=='0')
	  if(cpalette[1]==0)
	  {
	   free(cpalette);
	   cpalette=0;
	  }
	}
	if(docheckie[0])
	 FCEUI_SetVidSystem(1);
	if(docheckie[1])
	 FCEUI_SetGameGenie(1);

	if(docheckie[4])
	 netplay=2;
	else if(docheckie[5])
	 netplay=1;

        if(netplay)
         FCEUI_SetNetworkPlay(netplay,netskip);

        FCEUI_DisableSpriteLimitation(FCEUDemuloptions&1);
        FCEUI_SaveExtraDataUnderBase(FCEUDemuloptions&2);

        if(FCEUDsrendline<0 || FCEUDsrendline>239) FCEUDsrendline=0;
        if(FCEUDerendline<FCEUDsrendline || FCEUDerendline>239) FCEUDerendline=239;

        FCEUI_SetFirstRenderedLine(FCEUDsrendline);
        FCEUI_SetLastRenderedLine(FCEUDerendline);
        FCEUI_SetSoundVolume(soundvol);
	DriverInterface(DES_NTSCCOL,&ntsccol);

	for(x=0;x<4;x++)
	 if(!joy[x]) memset(joyBMap[x],0,4*sizeof(int));
	for(x=0;x<2;x++)
	{
	 int y;

         if(!inputa[x][0])
	  continue;

	 for(y=0;y<4;y++)	 
	 {
	  if(!strncmp(cortab[y],inputa[x],8))
	  {
	   InputType[x]=cortabi[y];
	   if(y==3)
	   {
	    powerpadside&=~(1<<x);
	    powerpadside|=((((inputa[x][8])-'a')&1)^1)<<x;
	   }
	  }
	 }
	}	
}

#include "usage.h"

int main(int argc, char *argv[])
{
	int ret;

        puts("\nStarting FCE Ultra "VERSION_STRING"...\n");
	vga_init();

	if(!(ret=FCEUI_Initialize()))
         return(1);
        GetBaseDirectory();
	CreateDirs();
        if(argc<=1) 
        {
         ShowUsage(argv[0]);
         return 1;
        }
        LoadConfig();
        DoArgs(argc-2,&argv[1]);
	if(cpalette)
	 LoadCPalette();
        if(soundo)
        {
	 int rate;
	 if(soundo==1)
	  soundo=48000;
	 rate=soundo;
         if(InitUNIXDSPSound(&rate,f8bit?0:1,sfragsize,snfrags))
          FCEUI_Sound(rate);
	 else
	  soundo=0;
        }

        if(!FCEUI_LoadGame(argv[argc-1]))
        {
         ret=0;
         goto dk;
        }

        if(!DriverInitialize())
        {
         ret=0;
         goto dk;
        }
	FCEUI_Emulate();

	dk:
	DriverKill();
        return(ret?0:1);
}

static int DriverInitialize(void)
{
   SetSignals((void *)CloseStuff);

   #ifdef LINUX
   InitJoystick();
   #endif

   if(!InitVideo()) return 0;
   if(!InitKeyboard()) return 0;
   InitOtherInput();
   return 1;
}

static void DriverKill(void)
{
 SaveConfig();
 SetSignals(SIG_IGN);
 #ifdef LINUX
 KillJoysticks();
 #endif
 if(winit&2)
  keyboard_close();
 KillVideo();
 if(soundo) 
  KillUNIXDSPSound(); 
 if(InputType[0]==DII_SETCONTROL_ZAPPER || InputType[1]==DII_SETCONTROL_ZAPPER)
  mouse_close();
}

void FCEUD_WriteSoundData(int32 *Buffer, int Count)
{
 WriteUNIXDSPSound(Buffer,Count,NoWaiting);
}
